searchNodes=[{"ref":"Data.Constructor.html","title":"Data.Constructor","type":"module","doc":""},{"ref":"Data.Constructor.html#struct/3","title":"Data.Constructor.struct/3","type":"function","doc":"Define and run a smart constructor on a Key-Value input, returning either well-defined structs or descriptive errors. The motto: parse, don&#39;t validate! Given a list of Data.Parser.KV.field_spec/2s, a module, and an input map or Keyword, create and run a parser which will either parse successfully and return an {:ok, %__MODULE__{}} struct, or fail and return an {:error, Error.t} with details about the parsing failure. Examples iex&gt; defmodule SensorReading do ...&gt; defstruct [:sensor_id, :microfrobs, :datetime] ...&gt; def new(input) do ...&gt; Data.Constructor.struct([ ...&gt; {:sensor_id, Data.Parser.BuiltIn.string()}, ...&gt; {:microfrobs, Data.Parser.BuiltIn.integer()}, ...&gt; {:datetime, Data.Parser.BuiltIn.datetime()}], ...&gt; __MODULE__, ...&gt; input) ...&gt; end ...&gt; end ...&gt; ...&gt; {:ok, reading} = SensorReading.new(sensor_id: &quot;1234-1234-1234&quot;, ...&gt; microfrobs: 23, ...&gt; datetime: ~U[2018-12-20 12:00:00Z]) ...&gt; ...&gt; ~U[2018-12-20 12:00:00Z] = reading.datetime ...&gt; 23 = reading.microfrobs ...&gt; &quot;1234-1234-1234&quot; = reading.sensor_id ...&gt; {:error, e} = SensorReading.new(%{&quot;sensor_id&quot; =&gt; nil, ...&gt; &quot;microfrobs&quot; =&gt; 23, ...&gt; &quot;datetime&quot; =&gt; &quot;2018-12-20 12:00:00Z&quot;}) ...&gt; :failed_to_parse_field = Error.reason(e) ...&gt; %{field: :sensor_id, input: %{&quot;datetime&quot; =&gt; &quot;2018-12-20 12:00:00Z&quot;, ...&gt; &quot;microfrobs&quot; =&gt; 23, ...&gt; &quot;sensor_id&quot; =&gt; nil}} = Error.details(e) ...&gt; {:just, inner_error} = Error.caused_by(e) ...&gt; Error.reason(inner_error) :not_a_string"},{"ref":"Data.Constructor.html#update/3","title":"Data.Constructor.update/3","type":"function","doc":"Define a smart update function based on a list of field specifications, the struct type to be updated, and a Keyword or map of input params. Given a list of Data.Parser.KV.field_spec/2s, a module (which defines a struct), and an input map or Keyword, create an {:ok, &amp;fun/1} tuple, or fail and return an {:error, Error.t} with details about the parsing failure. The &amp;fun/1 in the ok-tuple can be applied to any struct as defined by module, and will update it with the fields provided in the constructor input. The crucial advantage here is that the input parameters are validated according to the provided Data.Parser.KV.field_spec/2s, so that using the same list of field_specs for new/3 and update/3 will result in correct-by-construction data both from construction and after updates. Additionally, if any of the field_specs define a default: value, that value will be explicitly allowed in updates for that field, along with the specified type. Examples iex&gt; defmodule ReadingWComment do ...&gt; defstruct [:sensor_id, :microfrobs, :datetime, :comments] ...&gt; ...&gt; defp fields, do: [ ...&gt; {:sensor_id, Data.Parser.BuiltIn.string()}, ...&gt; {:microfrobs, Data.Parser.BuiltIn.integer()}, ...&gt; {:datetime, Data.Parser.BuiltIn.datetime()}, ...&gt; {:comments, Data.Parser.BuiltIn.string(), default: nil}] ...&gt; ...&gt; def new(input) do ...&gt; Data.Constructor.struct(fields(), __MODULE__, input) ...&gt; end ...&gt; ...&gt; def update(sensor_reading, input) do ...&gt; case Data.Constructor.update(fields(), __MODULE__, input) do ...&gt; {:ok, update_fun} -&gt; update_fun.(sensor_reading) ...&gt; {:error, e} -&gt; {:error, e} ...&gt; end ...&gt; end ...&gt; end ...&gt; ...&gt; {:ok, reading} = ReadingWComment.new(sensor_id: &quot;1234-1234-1234&quot;, ...&gt; microfrobs: 23, ...&gt; datetime: ~U[2018-12-20 12:00:00Z], ...&gt; comments: &quot;delete me later&quot;) ...&gt; &quot;delete me later&quot; = reading.comments ...&gt; ...&gt; ...&gt; {:ok, reading2} = ReadingWComment.update(reading, ...&gt; microfrobs: 25, ...&gt; datetime: ~U[2018-12-20 13:00:00Z]) ...&gt; ~U[2018-12-20 13:00:00Z] = reading2.datetime ...&gt; 25 = reading2.microfrobs ...&gt; ...&gt; ...&gt; {:ok, reading3} = ReadingWComment.update(reading2, ...&gt; comments: nil) ...&gt; nil = reading3.comments ...&gt; ...&gt; ...&gt; {:error, e} = ReadingWComment.update(reading3, ...&gt; microfrobs: [1,2,3]) ...&gt; :invalid_parameter = Error.reason(e) ...&gt; %{key: :microfrobs, value: [1,2,3]} = Error.details(e) ...&gt; ...&gt; ...&gt; {:error, e} = ReadingWComment.update(%{&quot;my&quot; =&gt; &quot;special&quot;, &quot;map&quot; =&gt; &quot;type&quot;}, ...&gt; microfrobs: 25, ...&gt; datetime: ~U[2018-12-20 13:00:00Z]) ...&gt; :struct_type_mismatch = Error.reason(e) ...&gt; Error.details(e) %{expecting: Data.ConstructorTest.ReadingWComment, got: %{&quot;map&quot; =&gt; &quot;type&quot;, &quot;my&quot; =&gt; &quot;special&quot;}}"},{"ref":"Data.Enum.html","title":"Data.Enum","type":"module","doc":"Functions which extend built-in Enum functionality."},{"ref":"Data.Enum.html#atomize_keys!/2","title":"Data.Enum.atomize_keys!/2","type":"function","doc":"Takes a key-value-type Enum (map, MapSet, associative list) with binary keys, and returns a Keyword, where all keys have been converted to atoms. The optional paramter :safe, which defaults to true, determines whether new atoms should be generated dynamically on-the-fly. If :safe is true and an unknown atom would have to be created, this function throws an ArgumentError. If :safe is false, then a new atom is created. Generating new atoms on-the-fly poses a security risk in prodution systems and should be avoided. Examples # note: the atom :hello does not exist in the VM iex&gt; Data.Enum.atomize_keys!(%{&quot;hello&quot; =&gt; &quot;world&quot;}) ** (ArgumentError) argument error # note: now, the atom does exist! iex&gt; :present; Data.Enum.atomize_keys!(%{&quot;present&quot; =&gt; &quot;yes!&quot;}) [present: &quot;yes!&quot;] # now, we&#39;ll generate a new atom iex&gt; Data.Enum.atomize_keys!(%{&quot;something_new&quot; =&gt; &quot;DANGER ZONE&quot;}, ...&gt; safe: false) [something_new: &quot;DANGER ZONE&quot;]"},{"ref":"Data.Parser.html","title":"Data.Parser","type":"module","doc":"Higher-order functions to create and modify parsers."},{"ref":"Data.Parser.html#kv/1","title":"Data.Parser.kv/1","type":"function","doc":"See Data.Parser.KV.new/1."},{"ref":"Data.Parser.html#list/1","title":"Data.Parser.list/1","type":"function","doc":"Takes a parser p and creates a parser that will successfully parse lists of values that all satisfy p. Specifically, the input: 1) Must be a list 2) p must parse successfully all elements in the input If this is the case, the output will be {:ok, list_of_parsed_values}. If not all values can be parsed with p, the result will be the original parse error, enriched with the field :failed_element in the error details. If the input is not a list, the domain error :not_a_list will be returned. Examples iex&gt; Data.Parser.list(Data.Parser.BuiltIn.integer()).([]) {:ok, []} iex&gt; Data.Parser.list(Data.Parser.BuiltIn.integer()).([1,2,3]) {:ok, [1, 2, 3]} iex&gt; {:error, e} = Data.Parser.list(Data.Parser.BuiltIn.integer()).(%{a: :b}) ...&gt; Error.reason(e) :not_a_list iex&gt; {:error, e} = Data.Parser.list(Data.Parser.BuiltIn.integer()).([1, :b, 3]) ...&gt; Error.reason(e) :not_an_integer ...&gt; Error.details(e) %{failed_element: :b}"},{"ref":"Data.Parser.html#maybe/1","title":"Data.Parser.maybe/1","type":"function","doc":"Takes a parser and transforms it so that it works &#39;inside&#39; Maybe.t values. If the original parser works on String.t(), the new one will work on Maybe.t(String.t()). Successful parses on just() values return {:ok, {:just, result_of_parse}}. Unsuccessful parses on just() values reutrn {:error, parse_error}. The parser will successfully return {:ok, :nothing} when applied to :nothing. Examples iex(2)&gt; Data.Parser.maybe( ...&gt; Data.Parser.predicate( &amp;String.valid?/1, :invalid)).({:just, &quot;good&quot;}) {:ok, {:just, &quot;good&quot;}} iex&gt; Data.Parser.maybe( ...&gt; Data.Parser.predicate( &amp;String.valid?/1, :invalid)).({:just, &#39;bad&#39;}) {:error, :invalid} iex&gt; Data.Parser.maybe( ...&gt; Data.Parser.predicate( &amp;String.valid?/1, :invalid)).(:nothing) {:ok, :nothing}"},{"ref":"Data.Parser.html#nonempty_list/1","title":"Data.Parser.nonempty_list/1","type":"function","doc":"Creates a parser that behaves exactly the same as the list/1 parser, except that it will return the domain error :empty_list if applied to an empty list. Examples iex&gt; Data.Parser.nonempty_list(Data.Parser.BuiltIn.integer()).([1, 2, 3]) {:ok, [1, 2, 3]} iex&gt; {:error, e} = Data.Parser.nonempty_list(Data.Parser.BuiltIn.integer()).([1, :b, 3]) ...&gt; Error.reason(e) :not_an_integer ...&gt; Error.details(e) %{failed_element: :b} iex&gt; {:error, e} = Data.Parser.nonempty_list(Data.Parser.BuiltIn.integer()).([]) ...&gt; Error.reason(e) :empty_list"},{"ref":"Data.Parser.html#one_of/2","title":"Data.Parser.one_of/2","type":"function","doc":"Takes a list of values, elements, and returns a parser that returns successfully if its input is present in elements. If the input is not a member of elements and default is a value, the parser fails with {:error, default}. If default is a unary function, the parser fails with {:error, default.(input)}. Examples iex&gt; Data.Parser.one_of([:he, :ne, :ar, :kr, :xe, :rn], &quot;not a noble gas&quot;).(:he) {:ok, :he} iex&gt; Data.Parser.one_of([:he, :ne, :ar, :kr, :xe, :rn], &quot;not a noble gas&quot;).(:n) {:error, &quot;not a noble gas&quot;} iex&gt; Data.Parser.one_of([:he, :ne, :ar, :kr, :xe, :rn], ...&gt; fn x -&gt; &quot;not a noble gas: \#{inspect x}&quot; end).(:o) {:error, &quot;not a noble gas: :o&quot;}"},{"ref":"Data.Parser.html#predicate/1","title":"Data.Parser.predicate/1","type":"function","doc":"Takes a boolean function p (the predicate), and returns a parser that parses successfully those values for which p is true. If the predicate returns false the parser will return a domain Error with the input value and the predicate functions listed in the error details. Examples iex&gt; {:error, e} = Data.Parser.predicate(&amp;String.valid?/1).(&#39;charlists are not ok&#39;) ...&gt; e.reason :predicate_not_satisfied ...&gt; e.details %{input: &#39;charlists are not ok&#39;, predicate: &amp;String.valid?/1} iex&gt; Data.Parser.predicate(&amp;String.valid?/1).(&quot;this is fine&quot;) {:ok, &quot;this is fine&quot;} iex&gt; Data.Parser.predicate(&amp;(&amp;1&lt;10)).(5) {:ok, 5} iex&gt; {:error, e} = Data.Parser.predicate(&amp;(&amp;1&lt;10)).(55) ...&gt; e.details.input 55"},{"ref":"Data.Parser.html#predicate/2","title":"Data.Parser.predicate/2","type":"function","doc":"Takes a boolean function p (the predicate) and a default value, and returns a parser that parses successfully those values for which p is true. If the predicate function applied to the input returns true, the parser wraps the input in an {:ok, input} tuple. If the predicate function returns false, and default is a value, the parser returns {:error, default} If the predicate returns false and default is a unary function, the parser returns {:error, default.(the_failed_input)}. Examples iex&gt; Data.Parser.predicate(&amp;String.valid?/1, &quot;invalid string&quot;).(&#39;charlists are not ok&#39;) {:error, &quot;invalid string&quot;} iex&gt; Data.Parser.predicate(&amp;String.valid?/1, &quot;invalid string&quot;).(&lt;&lt;&quot;neither are invalid utf sequences&quot;, 99999&gt;&gt;) {:error, &quot;invalid string&quot;} iex&gt; Data.Parser.predicate(&amp;String.valid?/1, &quot;invalid string&quot;).(&quot;this is fine&quot;) {:ok, &quot;this is fine&quot;} iex&gt; Data.Parser.predicate(&amp;String.valid?/1, fn x -&gt; &quot;the bad value is: \#{inspect x}&quot; end).(12345) {:error, &quot;the bad value is: 12345&quot;}"},{"ref":"Data.Parser.html#set/1","title":"Data.Parser.set/1","type":"function","doc":"Takes a parser p and creates a parser that will successfully parse sets of values that all satisfy p. Specifically, the input: 1) must be a MapSet 2) all elements of the input set must be parsed correctly by p If this is the case, the output will be {:ok, set_of_parsed_values}. If not all values can be parsed with p, the result will be the original parse error, enriched with the field :failed_element in the error details. If the input is not a MapSet, the domain error :not_a_set will be returned. Examples iex&gt; {:ok, s} = Data.Parser.set(Data.Parser.BuiltIn.integer()).(MapSet.new()) ...&gt; s #MapSet&lt;[]&gt; iex&gt; {:ok, s} = Data.Parser.set(Data.Parser.BuiltIn.integer()).(MapSet.new([1,2,3])) ...&gt; s #MapSet&lt;[1, 2, 3]&gt; iex&gt; {:error, e} = Data.Parser.set(Data.Parser.BuiltIn.integer()).(%{a: :b}) ...&gt; Error.reason(e) :not_a_set iex&gt; {:error, e} = Data.Parser.set(Data.Parser.BuiltIn.integer()).(MapSet.new([1, :b, 3])) ...&gt; Error.reason(e) :not_an_integer ...&gt; Error.details(e) %{failed_element: :b}"},{"ref":"Data.Parser.html#union/1","title":"Data.Parser.union/1","type":"function","doc":"Takes a list of parsers and creates a parser that returns the first successful parse result, or an error listing the parsers and the failed input. Examples iex&gt; Data.Parser.union( ...&gt; [Data.Parser.BuiltIn.integer(), ...&gt; Data.Parser.BuiltIn.boolean()]).(true) {:ok, true} iex&gt; Data.Parser.union( ...&gt; [Data.Parser.BuiltIn.integer(), ...&gt; Data.Parser.BuiltIn.boolean()]).(1) {:ok, 1} iex&gt; {:error, e} = Data.Parser.union( ...&gt; [Data.Parser.BuiltIn.integer(), ...&gt; Data.Parser.BuiltIn.boolean()]).(:atom) ...&gt; Error.reason(e) :no_parser_applies ...&gt; Error.details(e).input :atom"},{"ref":"Data.Parser.html#t:t/2","title":"Data.Parser.t/2","type":"type","doc":"A parser is a function that takes any value as input and produces a Result.t. More specifically, a parser(a,b) is a fuction that takes any input and returns {:ok, a} on a successful parse or {:error, b} if parsing failed."},{"ref":"Data.Parser.BuiltIn.html","title":"Data.Parser.BuiltIn","type":"module","doc":"Parsers for built-in Elixir data types."},{"ref":"Data.Parser.BuiltIn.html#atom/0","title":"Data.Parser.BuiltIn.atom/0","type":"function","doc":"Creates a parser that successfully parses atoms, and returns the domain error :not_an_atom for all other inputs. Examples iex&gt; Data.Parser.BuiltIn.atom().(:atom) {:ok, :atom} iex&gt; Data.Parser.BuiltIn.atom().(:other_atom) {:ok, :other_atom} iex&gt; {:error, e} = Data.Parser.BuiltIn.atom().(1.0) ...&gt; Error.reason(e) :not_an_atom iex&gt; {:error, e} = Data.Parser.BuiltIn.atom().([&quot;truth&quot;, &quot;or&quot;, &quot;dare&quot;]) ...&gt; Error.reason(e) :not_an_atom"},{"ref":"Data.Parser.BuiltIn.html#boolean/0","title":"Data.Parser.BuiltIn.boolean/0","type":"function","doc":"Creates a parser that successfully parses booleans, and returns the domain error :not_a_boolean for all other inputs. Examples iex&gt; Data.Parser.BuiltIn.boolean().(true) {:ok, true} iex&gt; Data.Parser.BuiltIn.boolean().(false) {:ok, false} iex&gt; {:error, e} = Data.Parser.BuiltIn.boolean().(1.0) ...&gt; Error.reason(e) :not_a_boolean iex&gt; {:error, e} = Data.Parser.BuiltIn.boolean().([:truth, :or, :dare]) ...&gt; Error.reason(e) :not_a_boolean"},{"ref":"Data.Parser.BuiltIn.html#date/0","title":"Data.Parser.BuiltIn.date/0","type":"function","doc":"Creates a parser that successfully parses Date.ts or String.t that represent legitimate Date.ts. Returns a domain error representing the parse failure if the string input cannot be parsed, and the domain error :not_a_date for all other inputs. Examples iex&gt; {:ok, d} = Data.Parser.BuiltIn.date().(~D[1999-12-31]) ...&gt; d ~D[1999-12-31] iex&gt; {:ok, d} = Data.Parser.BuiltIn.date().(&quot;1999-12-31&quot;) ...&gt; d ~D[1999-12-31] iex&gt; {:error, e} = Data.Parser.BuiltIn.date().(&quot;19991232&quot;) ...&gt; Error.reason(e) :invalid_format iex&gt; {:error, e} = Data.Parser.BuiltIn.date().(&quot;1999-12-32&quot;) ...&gt; Error.reason(e) :invalid_date iex&gt; {:error, e} = Data.Parser.BuiltIn.date().(123456789) ...&gt; Error.reason(e) :not_a_date"},{"ref":"Data.Parser.BuiltIn.html#datetime/0","title":"Data.Parser.BuiltIn.datetime/0","type":"function","doc":"Creates a parser that successfully parses DateTime.ts or String.t that represent legitimate DateTime.ts. Returns a domain error representing the parse failure if the string input cannot be parsed, and the domain error :not_a_datetime for all other inputs. Examples iex&gt; Data.Parser.BuiltIn.datetime().(~U[1999-12-31 23:59:59Z]) {:ok, ~U[1999-12-31 23:59:59Z]} iex&gt; Data.Parser.BuiltIn.datetime().(&quot;1999-12-31 23:59:59Z&quot;) {:ok, ~U[1999-12-31 23:59:59Z]} iex&gt; {:error, e} = Data.Parser.BuiltIn.datetime().(&quot;1999-12-32 23:59:59Z&quot;) ...&gt; Error.reason(e) :invalid_date iex&gt; {:error, e} = Data.Parser.BuiltIn.datetime().(&quot;1999-12-31 23:59:99Z&quot;) ...&gt; Error.reason(e) :invalid_time iex&gt; {:error, e} = Data.Parser.BuiltIn.datetime().(&quot;1999-12-31 23:59:59&quot;) ...&gt; Error.reason(e) :missing_offset iex&gt; {:error, e} = Data.Parser.BuiltIn.datetime().(123456789) ...&gt; Error.reason(e) :not_a_datetime"},{"ref":"Data.Parser.BuiltIn.html#float/0","title":"Data.Parser.BuiltIn.float/0","type":"function","doc":"Creates a parser that successfully parses floats, and returns the domain error :not_a_float for all other inputs. Examples iex&gt; Data.Parser.BuiltIn.float().(1.0) {:ok, 1.0} iex&gt; {:error, e} = Data.Parser.BuiltIn.float().(1) ...&gt; Error.reason(e) :not_a_float iex&gt; {:error, e} = Data.Parser.BuiltIn.float().(:hi) ...&gt; Error.reason(e) :not_a_float"},{"ref":"Data.Parser.BuiltIn.html#integer/0","title":"Data.Parser.BuiltIn.integer/0","type":"function","doc":"Creates a parser that successfully parses integers, and returns the domain error :not_an_integer for all other inputs. Examples iex&gt; Data.Parser.BuiltIn.integer().(1) {:ok, 1} iex&gt; {:error, e} = Data.Parser.BuiltIn.integer().(1.0) ...&gt; Error.reason(e) :not_an_integer iex&gt; {:error, e} = Data.Parser.BuiltIn.integer().(:hi) ...&gt; Error.reason(e) :not_an_integer"},{"ref":"Data.Parser.BuiltIn.html#naive_datetime/0","title":"Data.Parser.BuiltIn.naive_datetime/0","type":"function","doc":"Creates a parser that successfully parses NaiveDateTime.ts or String.t that represent legitimate NaiveDateTime.ts. Returns a domain error representing the parse failure if the string input cannot be parsed, and the domain error :not_a_naive_datetime for all other inputs. Examples iex&gt; Data.Parser.BuiltIn.naive_datetime.(~N[1999-12-31 23:59:59]) {:ok, ~N[1999-12-31 23:59:59]} iex&gt; Data.Parser.BuiltIn.naive_datetime.(&quot;1999-12-31 23:59:59&quot;) {:ok, ~N[1999-12-31 23:59:59]} iex&gt; {:error, e} = Data.Parser.BuiltIn.naive_datetime.(&quot;1999-12-32 23:59:59&quot;) ...&gt; Error.reason(e) :invalid_date iex&gt; {:error, e} = Data.Parser.BuiltIn.naive_datetime.(&quot;1999-12-31 23:59:99&quot;) ...&gt; Error.reason(e) :invalid_time iex&gt; {:error, e} = Data.Parser.BuiltIn.naive_datetime.(123456789) ...&gt; Error.reason(e) :not_a_naive_datetime"},{"ref":"Data.Parser.BuiltIn.html#null/0","title":"Data.Parser.BuiltIn.null/0","type":"function","doc":"Creates a parser that successfully parses only nil. Any other values, including false and the atom :nothing cause the parser to fail with a domain error of :not_nil Examples iex&gt; Data.Parser.BuiltIn.null().(nil) {:ok, nil} iex&gt; {:error, e} = Data.Parser.BuiltIn.null().(1.0) ...&gt; Error.reason(e) :not_nil iex&gt; {:error, e} = Data.Parser.BuiltIn.null().(false) ...&gt; Error.reason(e) :not_nil"},{"ref":"Data.Parser.BuiltIn.html#string/0","title":"Data.Parser.BuiltIn.string/0","type":"function","doc":"Creates a parser that succesfully parses String.ts (a.k.a binaries), and returns the domain error :not_a_string for all other inputs. Examples iex&gt; Data.Parser.BuiltIn.string().(&quot;hi&quot;) {:ok, &quot;hi&quot;} iex&gt; {:error, e} = Data.Parser.BuiltIn.string().(&#39;hi&#39;) ...&gt; Error.reason(e) :not_a_string iex&gt; {:error, e} = Data.Parser.BuiltIn.string().(:hi) ...&gt; Error.reason(e) :not_a_string"},{"ref":"Data.Parser.BuiltIn.html#string_of/1","title":"Data.Parser.BuiltIn.string_of/1","type":"function","doc":"Creates a parser that successfully parses strings representing either Integers or Floats. Returns a domain error detailing the parse failure on bad inputs. Look out! Partial results, such as that of Integer.parse(&quot;abc123&quot;), still count as errors! Examples iex&gt; Data.Parser.BuiltIn.string_of(Float).(&quot;1.1&quot;) {:ok, 1.1} iex&gt; {:error, e} = Data.Parser.BuiltIn.string_of(Float).(&quot;abc&quot;) ...&gt; Error.reason(e) :not_parseable_as_float ...&gt; Error.details(e) %{input: &quot;abc&quot;, native_parser_output: :error} iex&gt; Data.Parser.BuiltIn.string_of(Integer).(&quot;1234567890&quot;) {:ok, 1234567890} iex&gt; {:error, e} = Data.Parser.BuiltIn.string_of(Integer).(&quot;123abc&quot;) ...&gt; Error.reason(e) :not_parseable_as_integer ...&gt; Error.details(e) %{input: &quot;123abc&quot;, native_parser_output: {123, &quot;abc&quot;}} iex&gt; {:error, e} = Data.Parser.BuiltIn.string_of(Integer).([]) ...&gt; Error.reason(e) :not_a_string"},{"ref":"Data.Parser.KV.html","title":"Data.Parser.KV","type":"module","doc":"Creates parsers that accept KeyValue-style Enums as input. In particular, KV parsers work with: maps (e.g. %{&quot;hello&quot; =&gt; &quot;world&quot;}) Keyword.ts (e.g. [hello: &quot;world&quot;]) Lists of pairs (e.g. [{&quot;hello&quot;, &quot;world&quot;}]) KV parsers are higher-order parsers, and operate in roughly the same way as Data.Parser.list/1 or Data.Parser.set/1, but their definition is slightly more involved. A KV parser is created with a list of field_specs, where each field_spec defines what fields of the input to look at, and what parsers to run on them. Here are some examples of field_specs and their parsing behavior: {:username, Data.Parser.BuiltIn.string()} This spec says that the input must contain a :username field, and the value of that field must satisfy Data.Parser.BuiltIn.string/0. The output map will contain the key-value pair username: &quot;some string&quot;. If the field cannot be parsed successfully, the entire KV parser will return {:error, domain_error_with_details_on_parse_failure}. If the field is not present, the entire KV parser will return {:error, domain_error_with_details_about_field_not_found} {:birthday, Data.Parser.BuiltIn.date(), optional: true} This spec says that the input may contain a :birthday field. If the field does exist, it must satisfy Data.Parser.BuiltIn.date/0. If the field exists and parses successfully, the output map will contain the key-value pair birthday: {:just, ~D[1983-07-18]}. If the field does not exist, the output map will contain the key-value pair birthday: :nothing. If the field cannot be parsed successfully, the entire KV parser will return {:error, domain_error_with_parse_failure_details}. {:country, MyApp.country_parser(), default: &quot;Canada&quot;} This spec says that the input may contain a :country field, and if so, the value of that field must parse successfully with MyApp.country_parser/0. If the field exists and parses successfully, the output map will contain a key-value pair such as: country: &quot;Indonesia&quot;. If the field cannot be parsed successfully, the entire Constructor will return {:error, domain_error_with_details_on_parse_failure}. If the field does not exist, the default value will be used. In this case, the output will contain the key-value pair: country: &quot;Canada&quot; {:country, MyApp.country_parser(), nullable: true} This spec says that the input must contain a :country field, and if so, the value of that field must parse successfully with MyApp.country_parser/0 OR be equal to nil. If the field exists and parses successfully, the output map will contain a key-value pair such as: country: &quot;Indonesia&quot;. If the field cannot be parsed successfully, the entire Constructor will return {:error, domain_error_with_details_on_parse_failure}. However, if the value of the field is nil, this is treated as a successful parse. If the field does not exist, the parser will fail. Note that a field spec specified as nullable: true cannot also contain either the optional: true or default: x options. This is an illegal spec and will not be constructed, resulting in an {:error, domain_error} tuple, with :invalid_field_spec as the error reason. {:country, MyApp.country_parser(), from: :countryName} This spec says that the parser will use the data from :countryName in the input map. If the value under this key satisfies the MyApp.country_parser(), then the resulting value will be placed under the :country field. Note that the from keyname MUST always be specified as an atom, but it will be applied automatically to string keys. If the input contains both a string key and an atom key, the atom key will take priority. {:point, MyApp.point_parser(), recurse: true} Sometimes you want to run several different parsers on the same input map. For example, let&#39;s say your input looks like this: %{x: 12, y: -10, value: 34, name: &quot;exploding_barrel&quot;} But the data structure you want after parsing looks like this: %{point: %{x: 12, y: -10}, value: 34, name: &quot;exploding_barrel&quot;} And you have MyApp.point_parser() which accepts a map with :x and :y integer keys and constructs %{x: integer(), y: integer()}. You can define a field_spec with recurse: true and have that particular parser get run on its parent input map, not on the value of a field."},{"ref":"Data.Parser.KV.html#new/1","title":"Data.Parser.KV.new/1","type":"function","doc":"Given a list of field_specs, verify that all specs are well-formed and return {:ok, parser}, where parser will accept a map or Keyword input and apply the appropriate parsers to their corresponding fields. If the field_specs are not well-formed, return {:error, Error.t} with details about the invalid field_specs. Examples iex&gt; {:ok, p} = Data.Parser.KV.new([{:username, Data.Parser.BuiltIn.string()}]) ...&gt; p.(username: &quot;johndoe&quot;) {:ok, %{username: &quot;johndoe&quot;}} iex&gt; {:ok, p} = Data.Parser.KV.new([{:username, Data.Parser.BuiltIn.string()}]) ...&gt; p.(%{&quot;username&quot; =&gt; &quot;johndoe&quot;}) {:ok, %{username: &quot;johndoe&quot;}} iex&gt; {:error, e} = Data.Parser.KV.new([&quot;not a spec&quot;]) ...&gt; e.reason :invalid_field_spec ...&gt; e.details %{spec: &quot;not a spec&quot;} iex&gt; {:ok, p} = Data.Parser.KV.new([{:a, Data.Parser.BuiltIn.integer(), optional: true}]) ...&gt; p.(a: 1) {:ok, %{a: {:just, 1}}} iex&gt; {:ok, p} = Data.Parser.KV.new([{:a, Data.Parser.BuiltIn.integer(), optional: true}]) ...&gt; p.([]) {:ok, %{a: :nothing}} iex&gt; {:ok, p} = Data.Parser.KV.new([{:a, Data.Parser.BuiltIn.integer(), nullable: true}]) ...&gt; {:error, e} = p.([]) ...&gt; Error.reason(e) :field_not_found_in_input iex&gt; {:ok, p} = Data.Parser.KV.new([{:a, Data.Parser.BuiltIn.integer(), nullable: true}]) ...&gt; p.([a: nil]) {:ok, %{a: nil}} iex&gt; {:ok, p} = Data.Parser.KV.new([{:a, Data.Parser.BuiltIn.integer(), nullable: true}]) ...&gt; p.([a: 1]) {:ok, %{a: 1}} iex&gt; {:error, e} = Data.Parser.KV.new([{:a, Data.Parser.BuiltIn.integer(), nullable: true, default: nil}]) ...&gt; Error.reason(e) :invalid_field_spec iex&gt; {:error, e} = Data.Parser.KV.new([{:a, Data.Parser.BuiltIn.integer(), nullable: true, optional: true}]) ...&gt; Error.reason(e) :invalid_field_spec iex&gt; {:ok, p} = Data.Parser.KV.new([{:b, Data.Parser.BuiltIn.integer(), default: 0}]) ...&gt; p.([]) {:ok, %{b: 0}} iex&gt; {:ok, p} = Data.Parser.KV.new([{:b, Data.Parser.BuiltIn.integer(), default: 0}]) ...&gt; p.(b: 10) {:ok, %{b: 10}} iex&gt; {:ok, p} = Data.Parser.KV.new([{:b, Data.Parser.BuiltIn.integer(), default: 0}]) ...&gt; {:error, e} = p.(b: &quot;i am of the wrong type&quot;) ...&gt; Error.reason(e) :failed_to_parse_field ...&gt; {:just, inner_error} = Error.caused_by(e) ...&gt; Error.reason(inner_error) :not_an_integer iex&gt; {:ok, p} = Data.Parser.KV.new([{:a, Data.Parser.BuiltIn.integer(), from: :theAValue}]) ...&gt; p.(%{theAValue: 123}) {:ok, %{a: 123}} iex&gt; {:ok, p} = Data.Parser.KV.new([{:a, Data.Parser.BuiltIn.integer(), from: :aStringKey}]) ...&gt; p.(%{&quot;aStringKey&quot; =&gt; 1234}) {:ok, %{a: 1234}} iex&gt; {:ok, point} = Data.Parser.KV.new([{:x, Data.Parser.BuiltIn.integer()}, {:y, Data.Parser.BuiltIn.integer()}]) ...&gt; {:ok, item} = Data.Parser.KV.new([{:point, point, recurse: true}, {:value, Data.Parser.BuiltIn.integer()}]) ...&gt; item.(%{x: 1, y: -1, value: 34}) {:ok, %{value: 34, point: %{x: 1, y: -1}}} iex&gt; {:ok, point} = Data.Parser.KV.new([{:x, Data.Parser.BuiltIn.integer()}, {:y, Data.Parser.BuiltIn.integer()}]) ...&gt; {:ok, item} = Data.Parser.KV.new([{:point, point, recurse: true}, {:value, Data.Parser.BuiltIn.integer()}]) ...&gt; {:error, e} = item.(%{x: &quot;wrong&quot;, y: -1, value: 34}) ...&gt; {:just, e2} = e.caused_by ...&gt; e2.reason :failed_to_parse_field iex&gt; {:ok, point} = Data.Parser.KV.new([{:x, Data.Parser.BuiltIn.integer()}, {:y, Data.Parser.BuiltIn.integer()}]) ...&gt; {:ok, item} = Data.Parser.KV.new([{:point, point, recurse: true}, {:value, Data.Parser.BuiltIn.integer()}]) ...&gt; {:error, e} = item.(%{y: -1, value: 34}) ...&gt; {:just, e2} = e.caused_by ...&gt; e2.reason :field_not_found_in_input"},{"ref":"Data.Parser.KV.html#one/1","title":"Data.Parser.KV.one/1","type":"function","doc":"Given one field_spec, verify that it is well-formed and return {:ok, parser}, where parser will accept a map or Keyword input and attempt to parse that single field out of the map. Any field_spec with default semantics will be &#39;lifted&#39; to also accept the default value if present under the key. Any field_spec with optional semantics will be stripped of these, so that it can &#39;select&#39; only fields which exist. If the field_spec is not well-formed, return {:error, Error.t} with details about the invalid field_spec. Examples iex&gt; {:ok, p} = Data.Parser.KV.one({:username, Data.Parser.BuiltIn.string()}) ...&gt; p.(username: &quot;johndoe&quot;) {:ok, %{username: &quot;johndoe&quot;}} iex&gt; {:ok, p} = Data.Parser.KV.one({:username, Data.Parser.BuiltIn.string()}) ...&gt; p.(%{&quot;username&quot; =&gt; &quot;johndoe&quot;}) {:ok, %{username: &quot;johndoe&quot;}} iex&gt; {:ok, p} = Data.Parser.KV.one({:username, Data.Parser.BuiltIn.string(), default: nil}) ...&gt; p.(username: nil) {:ok, %{username: nil}} iex&gt; {:ok, p} = Data.Parser.KV.one({:username, Data.Parser.BuiltIn.string(), default: nil}) ...&gt; {:error, e} = p.(%{username: [1,2,3]}) ...&gt; Error.reason(e) :failed_to_parse_field iex&gt; {:ok, p} = Data.Parser.KV.one({:username, Data.Parser.BuiltIn.string(), default: nil}) ...&gt; {:error, e} = p.(%{&quot;a&quot; =&gt; &quot;b&quot;}) ...&gt; Error.reason(e) :field_not_found_in_input iex&gt; {:ok, p} = Data.Parser.KV.one({:username, Data.Parser.BuiltIn.string(), optiona: true}) ...&gt; {:error, e} = p.(%{&quot;a&quot; =&gt; &quot;b&quot;}) ...&gt; Error.reason(e) :field_not_found_in_input"},{"ref":"Data.Parser.KV.html#t:field/2","title":"Data.Parser.KV.field/2","type":"opaque","doc":"A structure representing a Data.Parser.t(a,b) lifted to operate on a KV."},{"ref":"Data.Parser.KV.html#t:field_name/0","title":"Data.Parser.KV.field_name/0","type":"type","doc":"KV parsers accept atom()s as key names, but will work on inputs where the keys are String.t()s as well."},{"ref":"Data.Parser.KV.html#t:field_opts/1","title":"Data.Parser.KV.field_opts/1","type":"type","doc":"Options to relax requirements on the fields. This is a list that consists of zero or one of the below options: {:optional, bool()} {:default, any} {:from, field_name()}"},{"ref":"Data.Parser.KV.html#t:field_spec/2","title":"Data.Parser.KV.field_spec/2","type":"type","doc":"A 2-tuple or 3-tuple describing the field to parse and parsing semantics. {field_name, parser} {field_name, parser, opts}"},{"ref":"Data.Parser.KV.html#t:input/0","title":"Data.Parser.KV.input/0","type":"type","doc":"KV parsers accept either a map or a Keyword.t as input."},{"ref":"readme.html","title":"Data","type":"extras","doc":"Data Extensions to Elixir data structures. Currently, this library contains 3 major components: Data.Enum — extensions to Enums. Data.Parser — parsers and parser combinators for BEAM data types. Data.Constructor — smart constructors (struct building w/ validation) based on Data.Parsers. For detailed documentation, see hex.pm."}]